@@ -12,11 +12,20 @@ import mmap
 import os.path
 import shutil
 import tempfile
-import thread
+import _thread as thread
 import threading
 import time
 import weakref
 
+KEY_SUFFIX = '.key'
+RAW_SUFFIX = '.raw'
+UTF_SUFFIX = '.utf'
+SER_SUFFIX = '.ser'
+FILE_SUFFIX = '.file'
+UNMAP_VALUE_SUFFIXES = (RAW_SUFFIX, UTF_SUFFIX, SER_SUFFIX)
+VALUE_SUFFIXES = (FILE_SUFFIX,) + UNMAP_VALUE_SUFFIXES
+ALL_SUFFIXES = VALUE_SUFFIXES + (KEY_SUFFIX,)
+
 _caches_mutex = threading.RLock()
 _caches = weakref.WeakKeyDictionary()
 
@@ -27,14 +36,14 @@ def _register_files(cache):
 
 def cachePurge(timeout = 0):
     with _caches_mutex:
-        caches = _caches.keys()
+        caches = list(_caches.keys())
 
     for cache in caches:
         cache.purge(timeout)
 
 def cacheClear():
     with _caches_mutex:
-        caches = _caches.keys()
+        caches = list(_caches.keys())
 
     for cache in caches:
         cache.clear()
@@ -75,6 +84,11 @@ def _touch(path):
     except OSError:
         pass
 
+def _nodelete(f):
+    f.delete = False
+    if hasattr(f, '_closer'):
+        f._closer.delete = False
+
 def _swap(source, dest, sizeback = None):
     cursize = 0
     if sizeback is not None:
@@ -159,6 +173,14 @@ def _clean(path, sizeback = None):
     if sizeback is not None:
         sizeback()
 
+def _clean_other_values(path, keep_suffix, sizeback = None):
+    for suffix in VALUE_SUFFIXES:
+        if suffix != keep_suffix:
+            _clean(path+suffix, sizeback)
+
+def json_dumpb(obj):
+    return json.dumps(obj).encode('ascii')
+
 class FilesCacheClient(base.BaseCacheClient):
     """
     This cache client will store itmes on a file hierarchy.
@@ -167,9 +189,11 @@ class FilesCacheClient(base.BaseCacheClient):
     it quite fast), but persistence is a possibility, and a common reason to use
     this client.
 
-    This client will behave specially when given file objects. Normal clients can't
-    or shouldn't store file objects, but this client has been specifically designed
-    for that task. File objects thus cached must contain a 'name' attribute that
+    This client will behave specially when given file, bytes, bytearray and memoryview
+    objects. Normal clients can't or shouldn't store file objects, but this client
+    has been specifically designed for that task.
+
+    File objects thus cached must contain a 'name' attribute that
     points to an absolute path to the file.
 
     When a file object is added to the client, it is copied in the most efficient
@@ -183,6 +207,8 @@ class FilesCacheClient(base.BaseCacheClient):
     an mmap object, depending on the value's size), and thus the cache will not be
     transparent, but it will be more efficient as it will share memory across processes.
 
+    Bytearray and memoryview objects are treated similarly.
+
     This cache is limited by size in bytes. The size includes book-keeping overhead,
     so be certain to account for serialized key sizes when specifying it.
 
@@ -200,8 +226,8 @@ class FilesCacheClient(base.BaseCacheClient):
 
     def __init__(self, size, basepath,
             failfast_size = 500, failfast_time = 0.25, counter_slots = 256,
-            key_pickler = json.dumps, value_pickler = None, value_unpickler = None, value_opener=None, checksum_key = None,
-            dirmode = 0700, filemode = 0400, mmap_raw = False,
+            key_pickler = json_dumpb, value_pickler = None, value_unpickler = None, value_opener=None, checksum_key = None,
+            dirmode = 0o700, filemode = 0o400, mmap_raw = False,
             sync_purge = None):
         """
         Params:
@@ -258,6 +284,8 @@ class FilesCacheClient(base.BaseCacheClient):
                 reached. To avoid the long pauses that could be generated by those forced purge cycles, this
                 is None (no sync purges) by default.
         """
+        if isinstance(checksum_key, str):
+            checksum_key = checksum_key.encode("utf8")
         self._failfast_cache = inproc.Cache(failfast_size)
         self.failfast_time = failfast_time
         self.basepath = basepath
@@ -277,12 +305,12 @@ class FilesCacheClient(base.BaseCacheClient):
 
         if value_pickler is None or value_unpickler is None:
             if checksum_key is None:
-                raise ValueError, "Must be given a checksum_key when using the default pickler"
+                raise ValueError("Must be given a checksum_key when using the default pickler")
             else:
                 self.value_pickler = functools.partial(sPickle.dump, checksum_key)
                 self.value_unpickler = functools.partial(sPickle.load, checksum_key)
         elif value_pickler is not None or value_unpickler is not None:
-            raise ValueError, "Must be given both pickler and unpickler when using custom picklers"
+            raise ValueError("Must be given both pickler and unpickler when using custom picklers")
         else:
             self.value_pickler = value_pickler
             self.value_unpickler = value_unpickler
@@ -357,7 +385,7 @@ class FilesCacheClient(base.BaseCacheClient):
         khash = hashlib.md5(key).hexdigest()
         return (khash[:2], khash[2:4], khash[4:6], khash[6:])
 
-    def _mktmp(self):
+    def _mktmp(self, **kw):
         tmpdir = os.path.join(self.basepath, "staging")
         if not os.path.exists(tmpdir):
             try:
@@ -365,10 +393,10 @@ class FilesCacheClient(base.BaseCacheClient):
             except:
                 if not os.path.exists(tmpdir):
                     raise
-        return tempfile.NamedTemporaryFile(dir=tmpdir)
+        return tempfile.NamedTemporaryFile(dir=tmpdir, **kw)
 
     @property
-    def async(self):
+    def is_async(self):
         return False
 
     @property
@@ -384,7 +412,7 @@ class FilesCacheClient(base.BaseCacheClient):
         kpath = self._mkpath(key)
 
         targetpath = os.path.join(self.basepath, *kpath)
-        keypath = targetpath + ".key"
+        keypath = targetpath + KEY_SUFFIX
         targetdir = os.path.join(self.basepath, *kpath[:-1])
 
         if not os.path.exists(targetdir):
@@ -398,7 +426,7 @@ class FilesCacheClient(base.BaseCacheClient):
         if os.path.exists(keypath):
             # Huh... compare
             if os.path.getsize(keypath) == len(key):
-                with open(keypath) as ekey:
+                with open(keypath, "rb") as ekey:
                     if ekey.read() == key:
                         reuse_keyfile = True
             if reuse_keyfile and not replace:
@@ -439,14 +467,13 @@ class FilesCacheClient(base.BaseCacheClient):
                     name = value.name
 
                     # Overwiting from unknown location, must take extra care
-                    _link(name, targetpath+'.file', self.size.__iadd__, self.filemode)
+                    _link(name, targetpath+FILE_SUFFIX, self.size.__iadd__, self.filemode)
                     if not reuse_keyfile:
                         _swap(keyfile.name, keypath, self.size.__iadd__)
-                        keyfile.delete = False
+                        _nodelete(keyfile)
 
                     # Succeeded, clean up other representations, if they exist, set ttl
-                    _clean(targetpath+'.raw', self.size.__isub__)
-                    _clean(targetpath+'.ser', self.size.__isub__)
+                    _clean_other_values(targetpath, FILE_SUFFIX, self.size.__isub__)
                     setTtl(keypath)
 
                     return True
@@ -456,41 +483,56 @@ class FilesCacheClient(base.BaseCacheClient):
                 pass
 
             # Else, must stuff the values into a file
-            if isinstance(value, (bytes,buffer)):
+            if isinstance(value, (bytes, memoryview, bytearray)):
                 # Quite easily... stuff it
                 with self._mktmp() as rawfile:
                     rawfile.write(value)
                     rawfile.flush()
 
-                    _swap(rawfile.name, targetpath+'.raw', self.size.__iadd__)
-                    rawfile.delete = False
+                    _swap(rawfile.name, targetpath+RAW_SUFFIX, self.size.__iadd__)
+                    _nodelete(rawfile)
                     if not reuse_keyfile:
                         _swap(keyfile.name, keypath, self.size.__iadd__)
-                        keyfile.delete = False
+                        _nodelete(keyfile)
 
                     # Succeeded, clean up other representations, if they exist
-                    _clean(targetpath+'.file', self.size.__isub__)
-                    _clean(targetpath+'.ser', self.size.__isub__)
+                    _clean_other_values(targetpath, RAW_SUFFIX, self.size.__isub__)
+                    setTtl(keypath)
+
+                return True
+            elif isinstance(value, (str,)):
+                # Quite easily... stuff it
+                with self._mktmp(mode='w+', encoding="utf8") as rawfile:
+                    rawfile.write(value)
+                    rawfile.flush()
+
+                    _swap(rawfile.name, targetpath+UTF_SUFFIX, self.size.__iadd__)
+                    _nodelete(rawfile)
+                    if not reuse_keyfile:
+                        _swap(keyfile.name, keypath, self.size.__iadd__)
+                        _nodelete(keyfile)
+
+                    # Succeeded, clean up other representations, if they exist
+                    _clean_other_values(targetpath, UTF_SUFFIX, self.size.__isub__)
                     setTtl(keypath)
 
                 return True
             else:
                 if not self.checksum_key:
-                    raise RuntimeError, "Cannot encode arbitrary objects without a checksum key"
+                    raise RuntimeError("Cannot encode arbitrary objects without a checksum key")
 
                 with self._mktmp() as rawfile:
                     self.value_pickler(value, rawfile, 2)
                     rawfile.flush()
 
-                    _swap(rawfile.name, targetpath+'.ser', self.size.__iadd__)
-                    rawfile.delete = False
+                    _swap(rawfile.name, targetpath+SER_SUFFIX, self.size.__iadd__)
+                    _nodelete(rawfile)
                     if not reuse_keyfile:
                         _swap(keyfile.name, keypath, self.size.__iadd__)
-                        keyfile.delete = False
+                        _nodelete(keyfile)
 
                     # Succeeded, clean up other representations, if they exist
-                    _clean(targetpath+'.file', self.size.__isub__)
-                    _clean(targetpath+'.raw', self.size.__isub__)
+                    _clean_other_values(targetpath, SER_SUFFIX, self.size.__isub__)
                     setTtl(keypath)
 
                 return True
@@ -521,7 +563,7 @@ class FilesCacheClient(base.BaseCacheClient):
         kpath = self._mkpath(key)
 
         targetpath = os.path.join(self.basepath, *kpath)
-        keypath = targetpath + ".key"
+        keypath = targetpath + KEY_SUFFIX
         targetdir = os.path.join(self.basepath, *kpath[:-1])
 
         if not os.path.exists(targetdir):
@@ -530,7 +572,7 @@ class FilesCacheClient(base.BaseCacheClient):
         if os.path.exists(keypath):
             # Huh... compare
             if os.path.getsize(keypath) == len(key):
-                with open(keypath) as ekey:
+                with open(keypath, "rb") as ekey:
                     if ekey.read() == key:
                         # Touch
                         now = time.time()
@@ -562,7 +604,7 @@ class FilesCacheClient(base.BaseCacheClient):
         kpath = self._mkpath(key)
         targetpath = os.path.join(self.basepath, *kpath)
 
-        for suffix in ('.file','.raw','.ser','.key'):
+        for suffix in ALL_SUFFIXES:
             _clean(targetpath+suffix, self.size.__isub__)
 
     def expire(self, key):
@@ -571,7 +613,7 @@ class FilesCacheClient(base.BaseCacheClient):
         targetpath = os.path.join(self.basepath, *kpath)
 
         # Touch all files, invalidating them immediately
-        for suffix in ('.file','.raw','.ser','.key'):
+        for suffix in ALL_SUFFIXES:
             try:
                 now = time.time()
                 os.utime(targetpath+suffix, (now,now))
@@ -579,11 +621,11 @@ class FilesCacheClient(base.BaseCacheClient):
                 pass
 
     def _getTtl(self, key, default = base.NONE, baseNONE = base.NONE, ttl_skip=None,
-            promote_callback = None, decode=True, return_stale=False):
+            promote_callback = None, decode=True, return_stale=False, _ninf=float('-inf')):
         key = self.key_pickler(key)
         kpath = self._mkpath(key)
         targetpath = os.path.join(self.basepath, *kpath)
-        keypath = targetpath + '.key'
+        keypath = targetpath + KEY_SUFFIX
 
         if not os.path.exists(keypath):
             return default, -1
@@ -595,10 +637,10 @@ class FilesCacheClient(base.BaseCacheClient):
             return default, -1
 
         # Check fail-fast cache before opening files and all taht
-        if self._failfast_cache.get(key) > (time.time() - self.failfast_time):
+        if self._failfast_cache.get(key, _ninf) > (time.time() - self.failfast_time):
             return default, -1
 
-        with open(keypath) as ekey:
+        with open(keypath, "rb") as ekey:
             if ekey.read() == key:
                 # Um... check the validity of the current value before going further
                 ttl = os.path.getmtime(keypath)
@@ -609,28 +651,31 @@ class FilesCacheClient(base.BaseCacheClient):
                 elif decode:
                     try:
                         _touch(keypath) # force atime update, in case it's a non-strict-atime mount
-                        if os.access(targetpath+'.file', os.R_OK):
-                            return self.value_opener(targetpath+'.file', 'rb') ,rttl
-                        elif os.access(targetpath+'.raw', os.R_OK):
-                            with open(targetpath+'.raw', 'rb') as rawfile:
+                        if os.access(targetpath+FILE_SUFFIX, os.R_OK):
+                            return self.value_opener(targetpath+FILE_SUFFIX, 'rb') ,rttl
+                        elif os.access(targetpath+RAW_SUFFIX, os.R_OK):
+                            with open(targetpath+RAW_SUFFIX, 'rb') as rawfile:
                                 rawfile.seek(0, os.SEEK_END)
                                 size = rawfile.tell()
                                 rawfile.seek(0)
                                 if size > MMAP_THRESHOLD:
                                     return mmap.mmap(rawfile.fileno(), size, access = mmap.ACCESS_READ), rttl
                                 elif size == 0:
-                                    return "", rttl
+                                    return b"", rttl
                                 else:
                                     return rawfile.read(), rttl
-                        elif self.checksum_key and os.access(targetpath+'.ser', os.R_OK):
-                            with open(targetpath+'.ser', 'rb') as rawfile:
+                        elif os.access(targetpath+UTF_SUFFIX, os.R_OK):
+                            with open(targetpath+UTF_SUFFIX, 'r', encoding="utf8") as rawfile:
+                                return rawfile.read(), rttl
+                        elif self.checksum_key and os.access(targetpath+SER_SUFFIX, os.R_OK):
+                            with open(targetpath+SER_SUFFIX, 'rb') as rawfile:
                                 try:
                                     return self.value_unpickler(rawfile), rttl
                                 except Exception:
                                     # Remove the broken entry from the cache
                                     logging.getLogger('chorde.files').warning(
                                         "Purging badly serialized entry from cache")
-                                    for suffix in ('.file','.raw','.ser','.key'):
+                                    for suffix in ALL_SUFFIXES:
                                         _clean(targetpath+suffix, self.size.__isub__)
                                     raise
                         else:
@@ -655,7 +700,7 @@ class FilesCacheClient(base.BaseCacheClient):
     def get(self, key, default=NONE, **kw):
         rv, ttl = self._getTtl(key, default, ttl_skip = 0, **kw)
         if ttl < 0 and default is NONE:
-            raise CacheMissError, key
+            raise CacheMissError(key)
         else:
             return rv
 
@@ -669,7 +714,7 @@ class FilesCacheClient(base.BaseCacheClient):
                 del self._failfast_cache[key]
             except:
                 pass
-            return ettl > ttl
+            return ttl is None or ettl > ttl
 
     def close(self):
         # Free up stuff
@@ -745,7 +790,7 @@ class FilesCacheClient(base.BaseCacheClient):
                 fullsize += getsize(join(basepath, fpath))
             except OSError:
                 pass
-            if fpath.endswith('.key'):
+            if fpath.endswith(KEY_SUFFIX):
                 bpath = fpath[:-4]
                 fpath = join(basepath, fpath)
                 bpath = join(basepath, bpath)
@@ -758,7 +803,7 @@ class FilesCacheClient(base.BaseCacheClient):
                     expirations.append(bpath)
                 elif deletions is not None:
                     itemsize = stats.st_size
-                    for suffix in ('.key', '.file','.raw','.ser'):
+                    for suffix in ALL_SUFFIXES:
                         try:
                             itemsize += os.path.getsize(bpath+suffix)
                         except OSError:
@@ -773,7 +818,7 @@ class FilesCacheClient(base.BaseCacheClient):
                         excess -= itemsize
                         heapq.heappush(deletions, delitem)
             else:
-                for suffix in ('.file','.raw','.ser'):
+                for suffix in VALUE_SUFFIXES:
                     if fpath.endswith(suffix):
                         bpath = fpath[:-len(suffix)]
                         break
@@ -782,7 +827,7 @@ class FilesCacheClient(base.BaseCacheClient):
                 if bpath is not None:
                     fpath = join(basepath, fpath)
                     bpath = join(basepath, bpath)
-                    if not exists(bpath+'.key'):
+                    if not exists(bpath+KEY_SUFFIX):
                         # Deleted entry
                         delta = []
                         _clean(fpath, delta.append)
@@ -806,7 +851,7 @@ class FilesCacheClient(base.BaseCacheClient):
         if expirations:
             for bpath in expirations:
                 delta = []
-                for suffix in ('.key','.file','.raw','.ser'):
+                for suffix in ALL_SUFFIXES:
                     _clean(bpath+suffix, delta.append)
                 if delta:
                     delta = sum(delta)
@@ -823,7 +868,7 @@ class FilesCacheClient(base.BaseCacheClient):
                 excess += delitem[2]
             for atime,bpath,itemsize in deletions:
                 delta = []
-                for suffix in ('.file','.raw','.ser','.key'):
+                for suffix in ALL_SUFFIXES:
                     _clean(bpath+suffix, delta.append)
                 if delta:
                     delta = sum(delta)
