@@ -3,9 +3,9 @@ import weakref
 import time
 import zmq
 import logging
-import Queue
+import queue
 import threading
-import thread
+import _thread as thread
 import random
 from abc import ABCMeta, abstractmethod
 
@@ -15,7 +15,7 @@ __ALL__ = (
     'ZMQIPSub',
 )
 
-FRAME_HEARTBEAT = "__HeyDude__"
+FRAME_HEARTBEAT = b"__HeyDude__"
 
 BROKER_PUB_HWM = 1000
 BROKER_REP_HWM = 1000
@@ -28,6 +28,10 @@ MIN_UPDATE_REPLY_FRAMES = 1
 MAX_UPDATE_REPLY_FRAMES = 1
 MAX_UPDATE_REPLY_FIRSTFRAME = 10
 
+ADDRINUSE_ERRNOS = (zmq.EADDRINUSE,)
+if hasattr(zmq, 'ENODEV'):
+    ADDRINUSE_ERRNOS += (zmq.ENODEV,)
+
 if hasattr(zmq, 'HWM'):
     # Has single HWM
     def set_hwm(sock, hwm):
@@ -76,12 +80,12 @@ class ZMQIPSub(BaseIPSub):
                 owner._pullsocket()
 
                 # Try a few times
-                for i in xrange(5):
+                for i in range(5):
                     try:
                         owner._bind()
                         break
                     except zmq.ZMQError as e:
-                        if e.errno in (zmq.EADDRINUSE, zmq.ENODEV):
+                        if e.errno in ADDRINUSE_ERRNOS:
                             # Not a transient error, shortcut to listener
                             return self.transition(ZMQIPSub.FSM.Listener)
                     except Exception as e:
@@ -185,19 +189,19 @@ class ZMQIPSub(BaseIPSub):
                             if what & POLLIN:
                                 recv_update_reply(listener_req)
                         # Then put incoming stuff on the queue
-                        for socket, what in activity.iteritems():
+                        for socket, what in activity.items():
                             if socket is pull:
                                 pack = pull_recv_multipart()
                                 if len_(pack) > 1:
                                     # ^ else Wakeup call, ignore
                                     put_nowait(pack)
-                                elif pack[0] == "tic":
+                                elif pack[0] == b"tic":
                                     owner._tic()
                                     tic_count = 100
                                 del pack
                             elif socket is listener_sub and what & POLLIN:
                                 recv_update(listener_sub)
-                except Queue.Full:
+                except queue.Full:
                     self.logger.error("While handling IPSub FSM pipe: Queue full, update lost")
                 except BootstrapNow:
                     self.transition(ZMQIPSub.FSM.Bootstrap)
@@ -248,7 +252,7 @@ class ZMQIPSub(BaseIPSub):
                 dict_ = dict
                 random_ = random.random
                 tic_count = 100
-                hb_period_base = owner.heartbeat_avg_period / 2
+                hb_period_base = owner.heartbeat_avg_period // 2
                 hb_period_spread = hb_period_base * 2
 
                 # Poll sockets
@@ -275,19 +279,19 @@ class ZMQIPSub(BaseIPSub):
                                 else:
                                     poller_unregister(broker_pub)
                         # Then put incoming stuff on the queue
-                        for socket, what in activity.iteritems():
+                        for socket, what in activity.items():
                             if socket is pull:
                                 pack = pull_recv_multipart()
                                 if len_(pack) > 1:
                                     # ^ else Wakeup call, ignore
                                     put_nowait(pack)
-                                elif pack[0] == "tic":
+                                elif pack[0] == b"tic":
                                     owner._tic()
                                     tic_count = 100
                                 del pack
                             elif socket is broker_rep and what & POLLIN:
                                 owner._handle_update_request(broker_rep)
-                except Queue.Full:
+                except queue.Full:
                     self.logger.error("While handling IPSub FSM pipe: Queue full, update lost")
                 except BootstrapNow:
                     self.transition(ZMQIPSub.FSM.Bootstrap)
@@ -306,7 +310,7 @@ class ZMQIPSub(BaseIPSub):
         super(ZMQIPSub, self).__init__()
 
         self.broker_addresses = broker_addresses
-        self.updates = Queue.Queue(INPROC_HWM)
+        self.updates = queue.Queue(INPROC_HWM)
         self.current_update = None
 
         self.listener_req = self.listener_sub = None
@@ -495,13 +499,13 @@ class ZMQIPSub(BaseIPSub):
             self.logger.error("IPSub FSM error: cannot send update when waiting for a reply, rebootstrapping")
             try:
                 self.updates.put_nowait(self.current_update)
-            except Queue.Full:
+            except queue.Full:
                 self.logger.error("While handling IPSub FSM error: Queue full, update lost")
             raise BootstrapNow
 
         try:
             update = self.updates.get_nowait()
-        except Queue.Empty:
+        except queue.Empty:
             # So what...
             return
 
@@ -561,7 +565,7 @@ class ZMQIPSub(BaseIPSub):
             try:
                 self.updates.put_nowait(update)
                 dropped = False
-            except Queue.Full:
+            except queue.Full:
                 dropped = True
 
             # Notify listeners
@@ -574,11 +578,13 @@ class ZMQIPSub(BaseIPSub):
                 socket.send(FRAME_UPDATE_OK)
 
     def publish(self, prefix, payload, timeout = None, _ident = thread.get_ident):
+        if isinstance(prefix, str):
+            prefix = prefix.encode("utf8")
         parts = [ prefix, self.identity ] + payload
         if _ident() == self.fsm_thread_id:
             try:
                 self.updates.put_nowait(parts)
-            except Queue.Full:
+            except queue.Full:
                 self.logger.error("While handling re-entrant IPSub publication: Queue full, update lost")
         else:
             push = self._pushsocket()
@@ -595,7 +601,7 @@ class ZMQIPSub(BaseIPSub):
                 # Don't block, mute means awaken and not catching up anyway
                 push = self._pushsocket()
                 if push.poll(1, zmq.POLLOUT):
-                    push.send("")
+                    push.send(b"")
             except zmq.ZMQError:
                 # Shit happens, probably not connected
                 pass
